#!/usr/bin/python
# gnerate_wrappers.py: Parse header and output libhybris binding code
# Adapted for libhybris from apkenv-wrapper-generator
#
# apkenv-wrapper-generator version:
# Copyright (C) 2012 Thomas Perl <m@thp.io>; 2012-10-19
#
# libhybris version:
# Copyright (C) 2013 Jolla Ltd.
# Contact: Thomas Perl <thomas.perl@jollamobile.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import re
import os
import sys

funcs = []

# Some special cases to to avoid having to parse the full C grammar
FIXED_MAPPINGS = {
        'GLint *': 'GLint *',
        'GLuint *': 'GLuint *',
        'GLsizei *': 'GLsizei *',
        'GLfloat eqn[4]': 'GLfloat[4]',
        'GLfixed eqn[4]': 'GLfixed[4]',
        'GLfixed mantissa[16]': 'GLfixed[16]',
        'GLint exponent[16]': 'GLint[16]',
        'GLfloat eqn[4]': 'GLfloat[4]',
        'GLvoid **': 'GLvoid **',
        'GLfloat *': 'GLfloat *',
        'GLfixed *': 'GLfixed *',
}

# Never generate wrappers for these functions
BLACKLISTED_FUNCS = [
]

def clean_arg(arg):
    arg = arg.replace('__restrict', '')
    arg = arg.replace('__const', '')
    arg = arg.replace('const', '')
    return arg.strip()

def mangle(name):
    return 'my_' + name

class Argument:
    def __init__(self, type_, name):
        self.type_ = type_
        self.name = name

class Function:
    def __init__(self, retval, name, args):
        self.retval = retval
        self.name = name
        self.raw_args = args
        self.parsing_error = None
        self.args = list(self._parse_args(args))

    def _parse_args(self, args):
        for arg in map(clean_arg, args.split(',')):
            arg = clean_arg(arg)
            xarg = re.match(r'^(.*?)([A-Za-z0-9_]+)$', arg)
            if not xarg:
                # Unknown argument
                if arg in FIXED_MAPPINGS:
                    yield Argument(FIXED_MAPPINGS[arg], '__undefined__')
                    continue
                self.parsing_error = 'Could not parse: ' + repr(arg)
                print self.parsing_error
                continue
            type_, name = xarg.groups()
            if type_ == '':
                type_ = name
                name = '__undefined__'
            yield Argument(type_, name)

if len(sys.argv) != 2:
    print >>sys.stderr, """
    Usage: %s glesv1_functions.h
    """ % (sys.argv[0],)
    sys.exit(1)

filename = sys.argv[1]

for line in open(filename):
    if line.startswith('/*'):
        continue
    retval, funcname, args = re.match(r'^(.+ [*]?)([A-Za-z0-9_]+)\((.*)\);\s*$', line).groups()
    retval, funcname, args = [x.strip() for x in (retval, funcname, args)]
    if funcname not in BLACKLISTED_FUNCS:
        funcs.append(Function(retval, funcname, args))

LIBRARY_NAME = 'glesv1_cm'

for function in funcs:
    args = [a.type_.strip() for a in function.args if a.type_.strip() not in ('', 'void')]
    if function.retval == 'void':
        print 'HYBRIS_IMPLEMENT_VOID_FUNCTION%d(%s, %s);' % (len(args), LIBRARY_NAME, ', '.join([function.name] + args))
    else:
        print 'HYBRIS_IMPLEMENT_FUNCTION%d(%s, %s, %s);' % (len(args), LIBRARY_NAME, function.retval, ', '.join([function.name] + args))




#!/usr/bin/python
#
# Generate wrapper macros: hybris/include/hybris/internal/binding.h
#
# Usage:
# python utils/generate_wrapper_macros.py >hybris/include/hybris/internal/binding.h
#
# Copyright (C) 2013 Jolla Ltd.
# Contact: Thomas Perl <thomas.perl@jollamobile.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# Maximum number of arguments to generate wrapper macros for
MAX_ARGS = 20

BEGIN, END = '{', '}'

AUTO_GENERATED_WARNING = """
/**
 *         XXX AUTO-GENERATED FILE XXX
 *
 * Do not edit this file directly, but update the templates in
 * utils/generate_wrapper_macros.py and run it again to build
 * an updated version of this header file:
 *
 *    python utils/generate_wrapper_macros.py > \\
 *       hybris/include/hybris/internal/binding.h
 *
 * If you need macros with more arguments, just customize the
 * MAX_ARGS variable in generate_wrapper_macros.py.
 *
 *         XXX AUTO-GENERATED FILE XXX
 **/
"""

print """
/**
 * Copyright (C) 2013 Simon Busch <morphis@gravedo.de>
 *               2012 Canonical Ltd
 *               2013 Jolla Ltd.
 *
 * Auto-generated via "generate_wrapper_macros.py"
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **/

#ifndef HYBRIS_BINDING_H_
#define HYBRIS_BINDING_H_

/* floating_point_abi.h defines FP_ATTRIB */
#include <hybris/common/floating_point_abi.h>

void *android_dlopen(const char *filename, int flag);
void *android_dlsym(void *name, const char *symbol);
int android_dlclose(void *handle);
const char *android_dlerror(void);
int android_dladdr(const void *addr, void *info);
"""

print AUTO_GENERATED_WARNING

print """
#define HYBRIS_DLSYSM(name, fptr, sym) \\
    if (!name##_handle) \\
        hybris_##name##_initialize(); \\
    if (*(fptr) == NULL) \\
    { \\
        *(fptr) = (void *) android_dlsym(name##_handle, sym); \\
    }

#define HYBRIS_LIBRARY_INITIALIZE(name, path) \\
    void *name##_handle; \\
    void hybris_##name##_initialize() \\
    { \\
        name##_handle = android_dlopen(path, RTLD_LAZY); \\
    }

"""

for count in range(MAX_ARGS):
    args = ['a%d' % (x+1) for x in range(count)]
    names = ['n%d' % (x+1) for x in range(count)]
    wrapper_signature = ', '.join(['name', 'return_type', 'symbol'] + args)
    signature = ', '.join(args)
    signature_with_names = ', '.join(' '.join(x) for x in zip(args, names))
    call_names = ', '.join(names)

    print """
#define HYBRIS_IMPLEMENT_FUNCTION{count}({wrapper_signature}) \\
    return_type symbol({signature_with_names}) \\
    {BEGIN} \\
        static return_type (*f)({signature}) FP_ATTRIB = NULL; \\
        HYBRIS_DLSYSM(name, &f, #symbol); \\
        return f({call_names}); \\
    {END}
""".format(**locals())

for count in range(MAX_ARGS):
    args = ['a%d' % (x+1) for x in range(count)]
    names = ['n%d' % (x+1) for x in range(count)]
    wrapper_signature = ', '.join(['name', 'symbol'] + args)
    signature = ', '.join(args)
    signature_with_names = ', '.join(' '.join(x) for x in zip(args, names))
    call_names = ', '.join(names)
    print """
#define HYBRIS_IMPLEMENT_VOID_FUNCTION{count}({wrapper_signature}) \\
    void symbol({signature_with_names}) \\
    {BEGIN} \\
        static void (*f)({signature}) FP_ATTRIB = NULL; \\
        HYBRIS_DLSYSM(name, &f, #symbol); \\
        f({call_names}); \\
    {END}
""".format(**locals())

# Print it again, so people wanting to append new macros will see it
print AUTO_GENERATED_WARNING

print """
#endif /* HYBRIS_BINDING_H_ */
"""




# Copyright (c) 2013 Adrian Negreanu <groleo@gmail.com>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

import os
import re
import subprocess

class LoadSymFiles(gdb.Command):
    """Add symbol files for files in /proc/pid/maps:  load-sym-file [symbols-directory] [lib-directory]"""
    """Usage example:
        start           # breakpoint 1
        set confirm off
        b dlfcn.c:56    # breakpoint 2, android_dlopen
        commands 2      # commands to run when breakpoint 2 is hit
        source ../utils/load_sym_files.py
        load-sym-files
        continue
        end             # end of commands for bkpt 2
    """
    def __init__(self):
        gdb.Command.__init__(self, "load-sym-files", gdb.COMMAND_FILES, gdb.COMPLETE_FILENAME, True)

    def invoke(self, arg, from_tty):
        arg_list = gdb.string_to_argv(arg)
        symdir = "/system/symbols"
        libdir = "/system/lib"

        if len(arg_list) == 1:
            symdir = arg_list[0]
        if not os.path.isdir(symdir):
            print "error: invalid symbol directory(%s)"%symdir
            print "usage: load-sym-files [symbols-dir] [lib-dir]"
            return

        if len(arg_list) == 2:
            libdir = arg_list[1]
        if not os.path.isdir(libdir):
            print "error: invalid library directory(%s)"%libdir
            print "usage: load-sym-files [symbols-dir] [lib-dir]"
            return
        try:
            pid = gdb.selected_inferior().pid
        except AttributeError:
            # in case gdb-version < 7.4
            # the array can have more than one element,
            # but it's good enough when debugging test_egl
            if len(gdb.inferiors()) == 1:
                pid = gdb.inferiors()[0].pid
            else:
                print "error: no gdb support for more than 1 inferior"
                return

        if pid == 0:
            print "error: debugee not started yet"
            return

        maps = open("/proc/%d/maps"%pid,"rb")
        for line in maps:
            # b7fc9000-b7fcf000 r-xp 00000000 08:01 1311443    /system/lib/liblog.so
            m = re.match("([0-9A-Fa-f]+)-[0-9A-Fa-f]+\s+r-xp.*(%s.*)"%libdir,line)
            if not m:
                continue

            start_addr = int(m.group(1),16)
            lib        = m.group(2)
            text_addr  = 0

            p = subprocess.Popen("objdump -h "+lib , shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE)
            for header in  p.stdout.read().split("\n"):
                #6 .text         00044ef7  00017f80  00017f80  00017f80  2**4
                t = re.match("\s*[0-9]+\s+\.text\s+([0-9A-Fa-f]+\s+){3}([0-9A-Fa-f]+)",header)
                if t:
                    text_addr = int(t.group(2),16)
                    break
            symfile = symdir + lib
            if os.path.isfile(symfile):
                gdb.execute("add-symbol-file %s 0x%X" % (symfile, start_addr+text_addr))


LoadSymFiles()
